import json
import typing
from crontab import CronTab, CronItem
from datetime import datetime, timedelta

from . import config
from .job import CronJob, _iterate_cronjobs


def current() -> CronTab:
    tabfile = config.tabfile()
    if tabfile:
        cron = CronTab(tabfile=tabfile)
    else:
        cron = CronTab(user=config.user())

    return cron


def _build_job_comment(job_id: str, module_name: str = None, run_type: str = None) -> str:
    job_comment_dict = {}
    instance = config.instance_name()
    if instance:
        job_comment_dict['instance'] = instance

    if module_name:
        job_comment_dict['module'] = module_name

    job_comment_dict['job_id'] = job_id

    if run_type:
        job_comment_dict['run_type'] = run_type

    return f'mara {json.dumps(job_comment_dict)}'


def get_job(cron: CronTab, job_id: str, module_name: str = None) -> typing.Optional[CronItem]:
    job_comment = _build_job_comment(job_id, module_name)

    for j in cron.find_comment(job_comment):
        return j


def generate() -> CronTab:
    cron = current()

    # iterate through defined cron jobs

    for module_name, cronjob in _iterate_cronjobs():
        job_comment = _build_job_comment(
                            job_id=cronjob.id,
                            module_name=(module_name if module_name != 'app' else None))

        job = None
        for j in cron.find_comment(job_comment):
            job = j
            job.set_command(cronjob.shell_command)
            break

        if not job:
            if not cronjob.time_pattern:
                continue
            job = cron.new(command=cronjob.shell_command, comment=job_comment)

        enabled = cronjob.enabled and config.enabled()

        if cronjob.time_pattern:
            job.setall(cronjob.time_pattern)
        else:
            # time pattern not set, keep the current time pattern and disable the cronjob
            enabled = False

        job.enable(enabled)

    return cron


def cronjob_time_pattern_from_datetime(datetime: datetime):
    """
    Generates a cron time pattern from a datetime object.

    Note:
        CRON does not support placing the year into the time pattern.
        We assume here that at least once a year the crontab is rewritten.
    """
    return f"{datetime.minute} {datetime.hour} {datetime.day} {datetime.month} *"


def append_single_execution_job(cronjob: CronJob):
    """Appends a new cron job for a single execution"""
    cron = current()

    job_comment_dict = {}
    instance = config.instance_name()
    if instance:
        job_comment_dict['instance'] = instance

    #if module_name:
    #    job_comment_dict['module'] = module_name

    job_comment_dict['job_id'] = cronjob.id
    job_comment_dict['type'] = 'single_execution'

    job = cron.new(command=cronjob.shell_command,
                   comment=_build_job_comment(cronjob.id, run_type='manual_scheduled'))

    # This job will run each year. We assume here that you will rewrite the cronjob task list before that time.
    time_pattern = str(cronjob_time_pattern_from_datetime(datetime.now() + timedelta(minutes=1)))
    job.setall(time_pattern)

    cron.write()



import sys
from warnings import warn
import click
from mara_cron import config

from . import crontab, job


@click.group()
def mara_cron():
    """Let you sync. mara jobs with your crontab"""
    pass


@mara_cron.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
@click.option('--module', default=None,
              help='The module name the cron job.')
def disable(job_id: str, module_name: str):
    """ disables a specific cronjob """
    cron = crontab.generate()

    job = crontab.get_job(cron, job_id=job_id, module_name=module_name)

    if not job:
        print(f'Could not find cronjob "{job_id}"', file=sys.stderr)
        sys.exit(1)

    job.enable(False)
    cron.write()
    sys.exit(0)


@mara_cron.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
@click.option('--module', default=None,
              help='The module name the cron job.')
def enable(job_id: str, module_name: str):
    """ enables a specific cronjob """
    if not config.enabled():
        print('WARNING: The global parameter mara_cron.config.enabled is not set. The job will be enabled nevertheless')

    cron = crontab.generate()

    job = crontab.get_job(cron, job_id=job_id, module_name=module_name)

    if not job:
        print(f'Could not find cronjob "{job_id}"', file=sys.stderr)
        sys.exit(1)

    job.enable(True)

    cron.write()
    sys.exit(0)


@mara_cron.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
def schedule_job(job_id: str):
    """ Schedules a job to run. """
    cronjob = job.find_job(job_id)
    if not cronjob:
        print(f'Could not find job with id "{job_id}"', file=sys.stderr)
        sys.exit(1)

    crontab.append_single_execution_job(cronjob)

    print(f'Job {job_id} is scheduled to run in less than 1 minute')
    sys.exit(0)


@mara_cron.command()
@click.option('--with-changes', default=False, is_flag=True,
              help='Lists the current crontab including the not written changes.')
def list_crontab(with_changes: bool):
    """ List the crontab content """

    if with_changes:
        cron = crontab.generate()
    else:
        cron = crontab.current()

    print(cron)


@mara_cron.command()
def write_crontab():
    """ Updates the current crontab """
    cron = crontab.generate()

    cron.write()


@mara_cron.command()
def clear_crontab():
    """ Removes all mara jobs from the crontab """
    cron = crontab.current()

    for job in cron:
        if job.comment.lstrip().startswith('mara'):
            print(f'Delted cron job: {job}')
            job.delete()

    cron.write()


# Old cli commands to be dropped in 1.0:

@click.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
@click.option('--module', default=None,
              help='The module name the cron job.')
def _disable(job_id: str, module_name: str):
    """ disables a specific cronjob """
    warn("CLI command `<app> mara_cron.disable` will be dropped in 5.0. Please use: `<app> mara-cron disable`")
    disable.callback(job_id=job_id, module_name=module_name)


@click.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
@click.option('--module', default=None,
              help='The module name the cron job.')
def _enable(job_id: str, module_name: str):
    """ enables a specific cronjob """
    warn("CLI command `<app> mara_cron.enable` will be dropped in 5.0. Please use: `<app> mara-cron enable`")
    enable.callback(job_id=job_id, module_name=module_name)


@click.command()
@click.option('--job-id', required=True,
              help='The id of of the cron job.')
def _schedule_job(job_id: str):
    """ Schedules a job to run. """
    warn("CLI command `<app> mara_cron.schedule_job` will be dropped in 5.0. Please use: `<app> mara-cron schedule_job`")
    schedule_job.callback(job_id=job_id)


@click.command()
@click.option('--with-changes', default=False, is_flag=True,
              help='Lists the current crontab including the not written changes.')
def _list_crontab(with_changes: bool):
    """ List the crontab content """
    warn("CLI command `<app> mara_cron.list_crontab` will be dropped in 5.0. Please use: `<app> mara-cron list_crontab`")
    list_crontab.callback(with_changes=with_changes)


@click.command("write_crontab")
def _write_crontab():
    """ Updates the current crontab """
    warn("CLI command `<app> mara_cron.write_crontab` will be dropped in 5.0. Please use: `<app> mara-cron write_crontab`")
    write_crontab.callback()


@click.command("clear_crontab")
def _clear_crontab():
    """ Removes all mara jobs from the crontab """
    warn("CLI command `<app> mara_cron.clear_crontab` will be dropped in 5.0. Please use: `<app> mara-cron clear_crontab`")
    clear_crontab.callback()



"""Mara admin views"""

import copy
import functools
import html
import sys
import typing

import flask
from mara_page import acl, navigation, response, _, bootstrap, xml
from . import config

from .job import CronJob

blueprint = flask.Blueprint('mara_cron', __name__, url_prefix='/mara-cron', static_folder='static')

acl_resource = acl.AclResource('Cron')


def _cronjob_modules(with_cronjobs=True):
    """Gathers all cron jobs from modules"""
    import inspect

    cronjob_modules = {}
    for name, module in copy.copy(sys.modules).items():
        if 'MARA_CRON_JOBS' in dir(module):
            cronjobs = getattr(module, 'MARA_CRON_JOBS')
            if isinstance(cronjobs, typing.Callable):
                cronjobs = cronjobs()
            assert (isinstance(cronjobs, typing.Iterable))
            cronjob_modules[f"{name}.crontab"] = {
                'doc': f'Module {name}' if name != 'app' else 'Main app',
                'cronjobs': {}
            }
            if with_cronjobs:
                for cronjob in cronjobs:
                    assert (isinstance(cronjob, CronJob))

                    cronjob_modules[f"{name}.crontab"]['cronjobs'][cronjob.id] = \
                        {
                            'doc': cronjob.description,
                            'time_pattern': cronjob.time_pattern,
                            'enabled': cronjob.enabled,
                            'command': cronjob.command
                        }

    return cronjob_modules


@blueprint.route('/crontab')
def crontab_page():
    import pprint

    # gather all cronjobs by package

    current_user_has_permission = acl.current_user_has_permission(acl_resource)

    return response.Response(
        html=[(bootstrap.card(id=module_name,
                              header_left=html.escape(module_name),
                              body=[_.p[_.em[html.escape(str(module['doc']))]],
                                    bootstrap.table(
                                        ['ID', 'Description', 'Status', 'Time pattern', 'Command', 'Actions'],
                                        [_.tr[
                                             _.td[html.escape(cronjob_id)],
                                             _.td[_.em[html.escape(cronjob['doc'])]],
                                             _.td[
                                                 _.span[
                                                     _.div(class_='fa fa-fw fa-play', style='color: green')[''],
                                                     'enabled'
                                                 ] if cronjob['enabled'] == True else _.span[
                                                     _.div(class_='fa fa-fw fa-stop', style='color: red')[''],
                                                     'disabled'
                                                 ]
                                             ],
                                             _.td[_.pre[html.escape(cronjob['time_pattern'] or '')]],
                                             _.td[
                                                 _.pre[html.escape(pprint.pformat(cronjob['command']))]
                                                 if current_user_has_permission
                                                 else acl.inline_permission_denied_message()
                                             ],
                                             _.td[
                                                 _.span[
                                                    bootstrap.button(url=flask.url_for('mara_cron.do_schedule_run', job_id=cronjob_id),
                                                                    label='Schedule run', icon='play',
                                                                    title='Schedule this task to run in less then 1 minute')
                                                        if config.allow_run_from_web_ui() else '',
                                                 ] if current_user_has_permission
                                                 else acl.inline_permission_denied_message()
                                             ]] for cronjob_id, cronjob in module['cronjobs'].items()])
                                    ]) if module['cronjobs'] else '')
              for module_name, module in sorted(_cronjob_modules().items())],
        title='Mara Crontab')


def package_cronjobs_navigation_entry():
    return navigation.NavigationEntry(
        label='Cron', icon='calendar', rank=100,
        description='Package cronjobs',
        uri_fn=lambda: flask.url_for('mara_cron.crontab_page'),
        children=[
            navigation.NavigationEntry(
                label=module_name, icon='list', description=module['doc'],
                uri_fn=lambda _module_name=module_name: flask.url_for('mara_cron.crontab_page',
                                                                      _anchor=_module_name))
            for module_name, module in sorted(_cronjob_modules(with_cronjobs=False).items())]
    )


@blueprint.route('/navigation-bar')
@functools.lru_cache(maxsize=None)
def navigation_bar() -> [str]:
    from . import app
    # The navigation sidebar is loaded asynchronously for better rendering experience
    def render_entries(entries: [navigation.NavigationEntry] = [], level: int = 1):
        def render_entry(entry: navigation.NavigationEntry, level: int = 1):
            attrs = {}
            if entry.children:
                attrs['onClick'] = 'toggleNavigationEntry(this)'
            else:
                attrs['href'] = entry.uri_fn()

            if entry.description:
                attrs.update({'title': entry.description, 'data-toggle': 'tooltip',
                              'data-container': 'body', 'data-placement': 'right'})
            return _.div(class_='mara-nav-entry level-' + str(level),
                         style='display:none' if level > 1 else '')[
                _.a(**attrs)[
                    _.div(class_='mara-nav-entry-icon fa fa-fw fa-' + entry.icon + (' fa-lg' if level == 1 else ''))[
                        ''] if entry.icon else '',
                    _.div(class_='mara-nav-entry-text')[entry.label.replace('_', '_<wbr>')],
                    _.div(class_='mara-caret fa fa-caret-down')[''] if entry.children else ''],
                render_entries(entry.children, level + 1)
            ]

        return [functools.partial(render_entry, level=level)(entry)
                for entry in sorted([entry for entry in entries if entry.visible], key=lambda x: x.rank)]

    return flask.Response(''.join(list(xml.render(render_entries(app.combine_navigation_entries().children)))))



"""Make the functionalities of this package auto-discoverable by mara-app"""
__version__ = '0.11'


def MARA_CONFIG_MODULES():
    from . import config
    return [config]


def MARA_FLASK_BLUEPRINTS():
    from . import ui, views
    return [views.blueprint]


def MARA_ACL_RESOURCES():
    from . import views
    return {'Cron': views.acl_resource}


def MARA_CLICK_COMMANDS():
    from . import cli
    return [
        cli.mara_cron,
        cli._enable,
        cli._disable,
        cli._schedule_job,
        cli._list_crontab,
        cli._write_crontab,
        cli._clear_crontab
    ]


def MARA_NAVIGATION_ENTRIES():
    from . import views
    return {'Cron Nav Entries Configs': views.package_cronjobs_navigation_entry()}



import typing as t


def enabled() -> bool:
    """
    A global parameter which lets you disable/activate all cron jobs

    Default value: False.
    """
    return False


def instance_name() -> t.Optional[str]:
    """
    Lets you define a name of your mara instance.

    Cron jobs will be executed per instance.
    """
    return None


def user() -> t.Union[bool, str]:
    """
    The crontab user:
    - you can return the name of crontab user
    - you can enter 'true' in case you want to get the crontab of the current user

    By default the current user is taken
    """
    return True


def tabfile() -> t.Optional[str]:
    """
    The crontab file. Use '/etc/crontab' in order to use

    By default not set. This will override the user() config.
    """
    return None


def allow_run_from_web_ui() -> bool:
    """When false, then it is not possible to start a cron job manually via the web UI"""
    return True


def log_path(): # -> t.Optional[t.Union[str, pathlib.Path]]:
    """
    The log path to which to log the command output.

    When the sign '$' is used in the path, shell quotation is disabled, otherwise the path name
    will be quoted.

    Note: You will have to ensure that the folder exist, otherwise logging will be skipped.

    It is recommended to use a different log path per instance e.g. '/var/log/mara/my_instance_name/'

    Possible options:
        Use a path to a folder e.g. '/var/log/mara'. You will have to make sure that the folder exists.
        Mara will automatically generate a log file named '$(date +%Y%m%d_%H%M%S).log',
        e.g. mara-cron_20211124_152800.log

        Alternatively you can define the name of the file as well by adding it to the path. You can add
        e.g. the date to a log file by using the following path: '/var/log/mara/mara_$(date +%Y%m%d).log'

        Each execution will append to the file. NOTE: When several jobs run at the same time, it will be
        maybe hard to determine to which execution the log line belongs.

    WARNING: Make sure that the user cron runs on has write access to the log_path. Otherwise the whole
             job will break with the error: Permission denied
    """
    return None


def default_job_max_retries() -> int:
    """How many times a cron job is retried when it fails by default"""
    return 0



import pytest
import subprocess
import pathlib

import mara_cron.config
from mara_app.monkey_patch import patch
from mara_cron.job import CronJob


ECHO_HELLO_WORLD_TEXT = 'Hello World!'
ECHO_HELLO_WORLD_COMMAND = f'echo "{ECHO_HELLO_WORLD_TEXT}"'
MARA_LOG_FILE_PATTERN = 'mara-cron_*.log'


@pytest.fixture
def job_echo_hello_world() -> CronJob:
    return CronJob(id='test_job',
                   description="",
                   default_time_pattern="00 1 * * *",
                   command=ECHO_HELLO_WORLD_COMMAND)


def test_log_path_does_not_exist(job_echo_hello_world: CronJob):
    patch(mara_cron.config.log_path)(lambda: 'tests/logs_folder_which_does_not_exist')
    job = job_echo_hello_world

    assert job.command == ECHO_HELLO_WORLD_COMMAND

    # since the log folder does not exist, the shell_command should be equal to the command
    assert job.shell_command == job.command


def test_execute_job_without_log(job_echo_hello_world: CronJob):
    patch(mara_cron.config.log_path)(lambda: None)
    job = job_echo_hello_world

    # validate internal fields
    assert job.command == ECHO_HELLO_WORLD_COMMAND
    assert job.shell_command == job.command

    # exec job
    (exitcode, stdout) = subprocess.getstatusoutput(job.shell_command)
    assert exitcode == 0
    assert stdout == ECHO_HELLO_WORLD_TEXT



def test_execute_job_with_logging(job_echo_hello_world: CronJob):
    patch(mara_cron.config.log_path)(lambda: 'tests/logs')
    job = job_echo_hello_world

    # Debug output
    print(f'job command: {job.command}')
    print(f'job shell command: {job.shell_command}')

    assert job.command == ECHO_HELLO_WORLD_COMMAND
    assert job.shell_command != job.command

    # exec job
    log_path = pathlib.Path(mara_cron.config.log_path())

    log_files = set(log_path.glob(MARA_LOG_FILE_PATTERN))

    (exitcode, stdout) = subprocess.getstatusoutput(job.shell_command)
    assert exitcode == 0
    assert not stdout

    new_log_files = set(log_path.glob(MARA_LOG_FILE_PATTERN))
    assert len(log_files) + 1 == len(new_log_files)

    log_file, = (new_log_files.difference(log_files))

    # test if we get a real file
    assert pathlib.Path(log_file).exists()

    # test if log file contains the stdout content
    found = False
    with open(log_file, 'r') as f:
        if ECHO_HELLO_WORLD_TEXT in f.read():
            found = True

    assert found



"""UI for manually scheduling a cron job"""

import flask

from mara_page import _, bootstrap, response, acl
from .. import views, job, crontab

@views.blueprint.route('/<path:job_id>/schedule-run')
@acl.require_permission(views.acl_resource)
def do_schedule_run(job_id: str):
    cronjob = job.find_job(job_id)
    if not cronjob:
        flask.abort(404, f'Job "{job_id}" not found')

    crontab.append_single_execution_job(cronjob)

    print(f'schedule {job_id}')

    return response.Response(
        title=f'Job scheduled',
        html=bootstrap.card(
            body=_.div[
                    _.p[
                        """The job is scheduled to run in less then 1 minute."""
                    ],
                    bootstrap.button(url='javascript:history.back()',
                                     label='Go to last page', icon='play',
                                     title='Returns to the last page')
                ]))



import copy
import os
import pathlib
import shlex
import sys
import typing

from . import config


class CronJob():
    """ A cron job configuration """
    def __init__(self, id: str, description: str, command: str,
                 default_time_pattern: str = None, default_enabled: bool = True,
                 max_retries: int = None):
        self.id = id
        self.description = description
        self.time_pattern = default_time_pattern
        self.command = command
        self.enabled = default_enabled
        self.max_retries = max_retries

    @property
    def shell_command(self):
        max_retries = self.max_retries
        if max_retries is None:
            max_retries = config.default_job_max_retries()

        if config.log_path():
            if max_retries:
                raise ValueError('ERROR: Logging of a cron job cannot be combined with max_retries in the current version of mara_cron!')
            log_path = pathlib.Path(config.log_path())

            log_full_path = None
            if log_path.is_file():
                log_full_path = str(log_path.absolute())
            elif log_path.is_dir():
                log_full_path = str((log_path / 'mara-cron_$(date "+%Y%m%d_%H%M%S").log').absolute())

            if log_full_path:
                log_command = f'{{ echo "MARA CRON JOB {self.id} START $(date)"; {self.command}; echo "MARA CRON JOB {self.id} END $(date)" ; }}'

                if '$' not in log_full_path:
                    log_full_path = shlex.quote(log_full_path)

                return f'{log_command} >> {log_full_path} 2>&1'

        if max_retries:
            return f"for i in {' '.join([str(2**(i+3)) for i in range(max_retries+1)])} 0; do ({self.command}) && break || sleep $i; done"

        return self.command


class MaraJob(CronJob):
    """ A configuration for a mara job"""
    def __init__(self, id: str, description: str, command: str, args: dict = None,
                 default_time_pattern: str = None, default_enabled = True,
                 max_retries: int = None):
        virtual_env_path = os.environ['VIRTUAL_ENV']
        if not virtual_env_path:
            raise Exception('Could not determine virtual environment path. VIRTUAL_ENV not set')

        mara_root_path = pathlib.Path(virtual_env_path).parent.resolve()

        job_command = f'cd {mara_root_path} ; . ./.venv/bin/activate ; flask {command}'

        if args:
            for param, value in args.items() if args else {}:
                job_command += f' {param}' \
                                + (f' {value}' if value and not isinstance(value, bool) else '')

        super().__init__(id=id, description=description, command=job_command,
                         default_time_pattern=default_time_pattern, default_enabled=default_enabled,
                         max_retries=max_retries)


class RunPipelineJob(MaraJob):
    """ A configuration for a job executing a mara pipeline"""
    def __init__(self, id: str, description: str, path: str = None, nodes: [str] = None, with_upstreams: bool = False,
                 default_time_pattern: str = None, default_enabled = True,
                 max_retries: int = None):
        """
        A job running a mara pipeline.
        """
        command = 'mara_pipelines.ui.run'
        args = {
            '--disable-colors': False
        }
        if path:
            args['--path'] = path
        if nodes:
            args['--nodes'] = ','.join(nodes)
        if with_upstreams:
            args['--with_upstreams'] = with_upstreams

        super().__init__(id=id, description=description, command=command, args=args,
                         default_time_pattern=default_time_pattern, default_enabled=default_enabled,
                         max_retries=max_retries)


def _iterate_cronjobs() -> typing.Dict[str, typing.List[CronJob]]:
    for module_name, module in copy.copy(sys.modules).items():
        if 'MARA_CRON_JOBS' in dir(module):
            cronjobs = getattr(module, 'MARA_CRON_JOBS')
            if isinstance(cronjobs, typing.Callable):
                cronjobs = cronjobs()
            assert (isinstance(cronjobs, typing.Iterable))
            for cronjob in cronjobs:
                assert (isinstance(cronjob, CronJob))
                yield module_name, cronjob


def find_job(id: str, module_name: str = None) -> CronJob:
    """
    Retrieves a job by the the job id
    Args:
        module_name: The name of the module
        id: The id of the job

    Returns:
        A CronJob object or None when the job was not found
    """
    if module_name:
        raise ValueError('Arg. module_name is not yet supported')

    for _, cronjob in _iterate_cronjobs():
        if cronjob.id == id:
            return cronjob

    return None
